"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const context_1 = require("../core/context");
const events_1 = require("./events");
const errors_1 = require("../errors");
var EngineState;
(function (EngineState) {
    EngineState[EngineState["AVAILABLE"] = 0] = "AVAILABLE";
    EngineState[EngineState["PROCESSED"] = 1] = "PROCESSED";
    EngineState[EngineState["ABORT"] = 2] = "ABORT";
})(EngineState || (EngineState = {}));
class Engine extends context_1.Context {
    constructor(id) {
        super(id, new events_1.EngineEvents());
        this.args = [];
        this.data = null;
        this.state = EngineState.AVAILABLE;
        this.onAbort = () => { };
    }
    clone() {
        const engine = new Engine(this.id);
        this.components.forEach(c => engine.register(c));
        return engine;
    }
    processStart() {
        if (this.state === EngineState.AVAILABLE) {
            this.state = EngineState.PROCESSED;
            return true;
        }
        if (this.state === EngineState.ABORT) {
            return false;
        }
        console.warn(`The process is busy and has not been restarted.
				Use abort() to force it to complete`);
        return false;
    }
    processDone() {
        const success = this.state !== EngineState.ABORT;
        this.state = EngineState.AVAILABLE;
        if (!success) {
            this.onAbort();
            this.onAbort = () => { };
        }
        return success;
    }
    abort() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise(resolve => {
                if (this.state === EngineState.PROCESSED) {
                    this.state = EngineState.ABORT;
                    this.onAbort = resolve;
                }
                else if (this.state === EngineState.ABORT) {
                    this.onAbort();
                    this.onAbort = resolve;
                }
                else {
                    resolve();
                }
            });
        });
    }
    lock(node) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise(resolve => {
                node.unlockPool = node.unlockPool || [];
                if (node.busy && !node.outputData) {
                    node.unlockPool.push(resolve);
                }
                else {
                    resolve();
                }
                node.busy = true;
            });
        });
    }
    unlock(node) {
        node.unlockPool.forEach(unlock => unlock());
        node.unlockPool = [];
        node.busy = false;
    }
    processWorker(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const inputData = yield this.extractInputData(node);
            const component = this.components.get(node.name);
            const outputData = new Map();
            try {
                yield component.worker(node, inputData, outputData, ...this.args);
            }
            catch (error) {
                this.abort();
                this.trigger('warn', error);
            }
            return outputData;
        });
    }
    processNode(node) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.state === EngineState.ABORT || !node) {
                return null;
            }
            yield this.lock(node);
            if (!node.outputData) {
                node.outputData = yield this.processWorker(node);
            }
            this.unlock(node);
            return node.outputData;
        });
    }
    extractInputData(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = new Map();
            for (const key of node.inputs.keys()) {
                const input = node.inputs.get(key);
                const connectionData = yield Promise.all(input.connections.map((connection) => __awaiter(this, void 0, void 0, function* () {
                    const previousNode = this.data.nodes.get(connection.nodeId);
                    const outputs = yield this.processNode(previousNode);
                    if (!outputs) {
                        this.abort();
                    }
                    else {
                        return outputs.get(connection.outputKey);
                    }
                })));
                data.set(key, connectionData);
            }
            return data;
        });
    }
    forwardProcess(node) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.state === EngineState.ABORT) {
                return null;
            }
            const processOutputs = [];
            for (const key of node.outputs.keys()) {
                processOutputs.push(new Promise((resolve) => __awaiter(this, void 0, void 0, function* () {
                    const output = node.outputs.get(key);
                    return resolve(yield Promise.all(output.connections.map((connection) => __awaiter(this, void 0, void 0, function* () {
                        const nextNode = this.data.nodes.get(connection.nodeId);
                        yield this.processNode(nextNode);
                        yield this.forwardProcess(nextNode);
                    }))));
                })));
            }
            return yield Promise.all(processOutputs);
        });
    }
    copy(data) {
        data = Object.assign({}, data);
        data.nodes = new Map(data.nodes);
        for (const key of data.nodes.keys()) {
            data.nodes.set(key, Object.assign({}, data.nodes.get(key)));
        }
        return data;
    }
    throwError(message, data = null) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.abort();
            this.trigger('error', { message, data });
            this.processDone();
            return 'error';
        });
    }
    validate(data) {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    processStartNode(id) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!id) {
                return;
            }
            const startNode = this.data.nodes.get(id);
            if (!startNode) {
                return yield this.throwError(errors_1.EngineError.StartNodeNotFound + id);
            }
            yield this.processNode(startNode);
            yield this.forwardProcess(startNode);
        });
    }
    processUnreachableNodes() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const node of this.data.nodes.values()) {
                if (node.outputData === undefined) {
                    yield this.processNode(node);
                    yield this.forwardProcess(node);
                }
            }
        });
    }
    process(data, startId = null, ...args) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.processStart()) {
                return;
            }
            if (!this.validate(data)) {
                return;
            }
            this.data = this.copy(data);
            this.args = args;
            yield this.processStartNode(startId);
            yield this.processUnreachableNodes();
            return this.processDone() ? 'success' : 'aborted';
        });
    }
}
exports.Engine = Engine;
//# sourceMappingURL=engine.js.map